import os
import sys
import json
import shutil

from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from cftool.misc import print_info
from cftool.misc import random_hash
from cftool.misc import update_dict
from cftool.misc import parse_config
from cftool.misc import print_warning
from cftool.misc import get_latest_workplace
from cftool.types import np_dict_type
from cftool.types import tensor_dict_type

from ..models import *
from .cv.pipeline import CVPipeline
from .ml.api import repeat_with
from .ml.api import RepeatResult
from .ml.pipeline import MLPipeline
from .ml.pipeline import MLCarefreePipeline
from .cv.models.diffusion import ldm as _ldm
from .cv.models.diffusion import ldm_vq as _ldm_vq
from .cv.models.diffusion import ldm_sd as _ldm_sd
from .cv.models.diffusion import ldm_sr as _ldm_sr
from .cv.models.diffusion import ldm_sd_anime as _ldm_sd_anime
from .cv.models.diffusion import ldm_semantic as _ldm_semantic
from .cv.models.diffusion import ldm_celeba_hq as _ldm_celeba_hq
from .cv.models.diffusion import ldm_inpainting as _ldm_inpainting
from .cv.models.translator import esr as _esr
from .cv.models.translator import esr_anime as _esr_anime
from ..data import MLData
from ..data import IMLData
from ..data import CVDataModule
from ..data import MLCarefreeData
from ..types import data_type
from ..types import configs_type
from ..types import general_config_type
from ..types import sample_weights_type
from ..types import states_callback_type
from ..trainer import get_sorted_checkpoints
from ..pipeline import DLPipeline
from ..pipeline import ModelSoupConfigs
from ..protocol import loss_dict
from ..protocol import metric_dict
from ..protocol import ILoss
from ..protocol import IDLModel
from ..protocol import _IMetric
from ..protocol import IDataLoader
from ..constants import SCORES_FILE
from ..constants import DEFAULT_ZOO_TAG
from ..constants import CHECKPOINTS_FOLDER
from ..zoo.core import _parse_config
from ..zoo.core import configs_root
from ..zoo.core import DLZoo
from ..dist.ml import Experiment
from ..misc.toolkit import inject_debug
from ..misc.toolkit import download_model
from ..models.protocols.ml import ml_core_dict


# dl


def make(name: str, *, config: general_config_type = None) -> DLPipeline:
    m = DLPipeline.make(name, parse_config(config))
    assert isinstance(m, DLPipeline)
    return m


def run_ddp(path: str, cuda_list: List[Union[int, str]], **kwargs: Any) -> None:
    def _convert_config() -> str:
        return " ".join([f"--{k}={v}" for k, v in kwargs.items()])

    os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, cuda_list))
    kwargs["nproc_per_node"] = len(cuda_list)
    prefix = f"{sys.executable} -m torch.distributed.run "
    os.system(f"{prefix}{_convert_config()} {path}")


def run_multiple(
    path: str,
    model_name: str,
    cuda_list: Optional[List[Union[int, str]]],
    *,
    num_jobs: int = 1,
    num_multiple: int = 5,
    workplace: str = "_multiple",
    sequential: Optional[bool] = None,
    resource_config: Optional[Dict[str, Any]] = None,
    is_fix: bool = False,
) -> None:
    if os.path.isdir(workplace) and not is_fix:
        print_warning(f"'{workplace}' already exists, it will be erased")
        shutil.rmtree(workplace)
    if sequential is None:
        sequential = num_jobs <= 1

    def is_buggy(i_: int) -> bool:
        i_workplace = os.path.join(workplace, model_name, str(i_))
        i_latest_workplace = get_latest_workplace(i_workplace)
        if i_latest_workplace is None:
            return True
        checkpoint_folder = os.path.join(i_latest_workplace, CHECKPOINTS_FOLDER)
        if not os.path.isfile(os.path.join(checkpoint_folder, SCORES_FILE)):
            return True
        if not get_sorted_checkpoints(checkpoint_folder):
            return True
        return False

    if sequential:
        for i in range(num_multiple):
            if is_fix and not is_buggy(i):
                continue
            cuda = "" if cuda_list is None else str(cuda_list[0])
            os.environ["CUDA_VISIBLE_DEVICES"] = cuda
            os.system(f"{sys.executable} {path}")
    else:
        if num_jobs <= 1:
            print_warning(
                "we suggest setting `sequential` "
                f"to True when `num_jobs` is {num_jobs}"
            )
        # generate temporary runtime script
        with open(path, "r") as rf:
            original_scripts = rf.read()
        folder = os.path.split(os.path.abspath(path))[0]
        tmp_path = os.path.join(folder, f"{random_hash()}.py")
        with open(tmp_path, "w") as wf:
            wf.write(
                f"""
import os
from cflearn.misc.toolkit import _set_environ_workplace
from cflearn.dist.ml.runs._utils import get_info

info = get_info(requires_data=False)
os.environ["CUDA_VISIBLE_DEVICES"] = str(info.meta["cuda"])
_set_environ_workplace(info.meta["workplace"])

{original_scripts}
"""
            )
        print_info(f"`{tmp_path}` will be executed")
        # construct & execute an Experiment
        cudas = None if cuda_list is None else list(map(int, cuda_list))
        experiment = Experiment(
            num_jobs=num_jobs,
            available_cuda_list=cudas,
            resource_config=resource_config,
        )
        for i in range(num_multiple):
            if is_fix and not is_buggy(i):
                continue
            experiment.add_task(
                model=model_name,
                run_command=f"{sys.executable} {tmp_path}",
                root_workplace=workplace,
            )
        experiment.run_tasks(use_tqdm=False)
        os.remove(tmp_path)


def pack(
    workplace: str,
    *,
    step: Optional[str] = None,
    config_bundle_callback: Optional[Callable[[Dict[str, Any]], Any]] = None,
    pack_folder: Optional[str] = None,
    cuda: Optional[Union[int, str]] = None,
    compress: bool = True,
    # model soup
    model_soup_loader: Optional[IDataLoader] = None,
    model_soup_metric_names: Optional[Union[str, List[str]]] = None,
    model_soup_metric_configs: configs_type = None,
    model_soup_metric_weights: Optional[Dict[str, float]] = None,
    model_soup_valid_portion: float = 1.0,
    model_soup_strategy: str = "greedy",
    model_soup_states_callback: states_callback_type = None,
    model_soup_verbose: bool = True,
) -> str:
    cls = DLPipeline.get_base(workplace)
    if model_soup_loader is None or model_soup_metric_names is None:
        model_soup_configs = None
    else:
        model_soup_configs = ModelSoupConfigs(
            model_soup_loader,
            model_soup_metric_names,
            model_soup_metric_configs,
            model_soup_metric_weights,
            model_soup_valid_portion,
            model_soup_strategy,
            model_soup_states_callback,
            model_soup_verbose,
        )
    return cls.pack(
        workplace,
        step=step,
        config_bundle_callback=config_bundle_callback,
        pack_folder=pack_folder,
        cuda=cuda,
        compress=compress,
        model_soup_configs=model_soup_configs,
    )


def load(
    export_folder: str,
    *,
    cuda: Optional[Union[int, str]] = None,
    compress: bool = True,
    states_callback: states_callback_type = None,
    pre_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    post_callback: Optional[Callable[[DLPipeline, Dict[str, Any]], None]] = None,
    strict: bool = True,
) -> DLPipeline:
    return DLPipeline.load(
        export_folder,
        cuda=cuda,
        compress=compress,
        states_callback=states_callback,
        pre_callback=pre_callback,
        post_callback=post_callback,
        strict=strict,
    )


def pack_onnx(
    workplace: str,
    export_folder: str,
    dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
    *,
    input_sample: tensor_dict_type,
    step: Optional[str] = None,
    config_bundle_callback: Optional[Callable[[Dict[str, Any]], Any]] = None,
    pack_folder: Optional[str] = None,
    states_callback: states_callback_type = None,
    pre_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    post_callback: Optional[Callable[[DLPipeline, Dict[str, Any]], None]] = None,
    onnx_file: str = "model.onnx",
    opset: int = 11,
    simplify: bool = True,
    num_samples: Optional[int] = None,
    verbose: bool = True,
    **kwargs: Any,
) -> DLPipeline:
    cls = DLPipeline.get_base(workplace)
    return cls.pack_onnx(
        workplace,
        export_folder,
        dynamic_axes,
        step=step,
        config_bundle_callback=config_bundle_callback,
        pack_folder=pack_folder,
        states_callback=states_callback,
        pre_callback=pre_callback,
        post_callback=post_callback,
        onnx_file=onnx_file,
        opset=opset,
        simplify=simplify,
        input_sample=input_sample,
        num_samples=num_samples,
        verbose=verbose,
        **kwargs,
    )


def from_json(d: Union[str, Dict[str, Any]]) -> DLPipeline:
    return DLPipeline.from_json(d)


def make_model(name: str, **kwargs: Any) -> IDLModel:
    return IDLModel.make(name, kwargs)


def make_loss(name: str, **kwargs: Any) -> ILoss:
    return ILoss.make(name, kwargs)


def make_metric(name: str, **kwargs: Any) -> _IMetric:
    return _IMetric.make(name, kwargs)


class ModelItem(NamedTuple):
    name: str
    requirements: Dict[str, Any]


def model_zoo(*, verbose: bool = False) -> List[ModelItem]:
    def _squeeze_requirements(req: Dict[str, Any], d: Dict[str, Any]) -> None:
        for k, v in req.items():
            kd = d.get(k)
            if kd is None:
                continue
            if isinstance(v, dict):
                _squeeze_requirements(v, kd)
                continue
            assert isinstance(v, list)
            pop_indices = []
            for i, vv in enumerate(v):
                if vv in kd:
                    pop_indices.append(i)
            for i in pop_indices[::-1]:
                v.pop(i)

    models = []
    for task in sorted(os.listdir(configs_root)):
        if task == "common_":
            continue
        task_folder = os.path.join(configs_root, task)
        for model in sorted(os.listdir(task_folder)):
            model_folder = os.path.join(task_folder, model)
            for config_file in sorted(os.listdir(model_folder)):
                config_path = os.path.join(model_folder, config_file)
                d = _parse_config(config_path)
                requirements = d.pop("__requires__", {})
                _squeeze_requirements(requirements, d)
                tag = os.path.splitext(config_file)[0]
                name = f"{task}/{model}"
                if tag != DEFAULT_ZOO_TAG:
                    name = f"{name}.{tag}"
                models.append(ModelItem(name, requirements))
    if verbose:

        def _stringify_item(item: ModelItem) -> str:
            return f"{item.name:>{span}s}   |   {json.dumps(item.requirements)}"

        span = 42
        print(
            "\n".join(
                [
                    "=" * 120,
                    f"{'Names':>{span}s}   |   Requirements",
                    "-" * 120,
                    "\n".join(map(_stringify_item, models)),
                    "-" * 120,
                ]
            )
        )
    return models


def from_zoo(
    model: str,
    *,
    build: bool = True,
    return_model: bool = False,
    **kwargs: Any,
) -> Union[IDLModel, DLPipeline]:
    if return_model and not build:
        raise ValueError("`build` should be True when `return_model` is True")
    kwargs["build"] = build
    fn = DLZoo.load_model if return_model else DLZoo.load_pipeline
    return fn(model, **kwargs)  # type: ignore


def supported_losses() -> List[str]:
    return sorted(loss_dict)


def supported_metrics() -> List[str]:
    return sorted(metric_dict)


# ml


def _make_ml_data(
    x_train: Union[data_type, IMLData],
    y_train: data_type = None,
    x_valid: data_type = None,
    y_valid: data_type = None,
    train_others: Optional[np_dict_type] = None,
    valid_others: Optional[np_dict_type] = None,
    carefree: bool = False,
    is_classification: Optional[bool] = None,
    data_config: Optional[Dict[str, Any]] = None,
    cf_data_config: Optional[Dict[str, Any]] = None,
) -> IMLData:
    if isinstance(x_train, IMLData):
        x_train.is_classification = is_classification
        return x_train
    data_kwargs: Dict[str, Any] = {
        "is_classification": is_classification,
        "train_others": train_others,
        "valid_others": valid_others,
    }
    if carefree:
        data_kwargs["cf_data_config"] = cf_data_config
    update_dict(data_config or {}, data_kwargs)
    args = x_train, y_train, x_valid, y_valid
    fn = MLCarefreeData.make_with if carefree else MLData
    return fn(*args, **data_kwargs)  # type: ignore


def fit_ml(
    x_train: Union[data_type, IMLData],
    y_train: data_type = None,
    x_valid: data_type = None,
    y_valid: data_type = None,
    *,
    train_others: Optional[np_dict_type] = None,
    valid_others: Optional[np_dict_type] = None,
    # data
    carefree: bool = False,
    is_classification: Optional[bool] = None,
    data_config: Optional[Dict[str, Any]] = None,
    cf_data_config: Optional[Dict[str, Any]] = None,
    # pipeline
    core_name: str = "fcnn",
    core_config: Optional[Dict[str, Any]] = None,
    input_dim: Optional[int] = None,
    output_dim: Optional[int] = None,
    loss_name: str = "auto",
    loss_config: Optional[Dict[str, Any]] = None,
    # encoder
    only_categorical: bool = False,
    encoder_config: Optional[Dict[str, Any]] = None,
    encoding_settings: Optional[Dict[int, Dict[str, Any]]] = None,
    # trainer
    state_config: Optional[Dict[str, Any]] = None,
    num_epoch: int = 40,
    max_epoch: int = 1000,
    fixed_epoch: Optional[int] = None,
    fixed_steps: Optional[int] = None,
    log_steps: Optional[int] = None,
    valid_portion: float = 1.0,
    amp: bool = False,
    clip_norm: float = 0.0,
    cudnn_benchmark: bool = False,
    metric_names: Optional[Union[str, List[str]]] = None,
    metric_configs: configs_type = None,
    metric_weights: Optional[Dict[str, float]] = None,
    use_losses_as_metrics: Optional[bool] = None,
    loss_metrics_weights: Optional[Dict[str, float]] = None,
    recompute_train_losses_in_eval: bool = True,
    monitor_names: Optional[Union[str, List[str]]] = None,
    monitor_configs: Optional[Dict[str, Any]] = None,
    callback_names: Optional[Union[str, List[str]]] = None,
    callback_configs: Optional[Dict[str, Any]] = None,
    lr: Optional[float] = None,
    optimizer_name: Optional[str] = None,
    scheduler_name: Optional[str] = None,
    optimizer_config: Optional[Dict[str, Any]] = None,
    scheduler_config: Optional[Dict[str, Any]] = None,
    optimizer_settings: Optional[Dict[str, Dict[str, Any]]] = None,
    use_zero: bool = False,
    workplace: str = "_logs",
    finetune_config: Optional[Dict[str, Any]] = None,
    tqdm_settings: Optional[Dict[str, Any]] = None,
    # misc
    pre_process_batch: bool = True,
    debug: bool = False,
    # fit
    sample_weights: sample_weights_type = None,
    cuda: Optional[Union[int, str]] = None,
) -> MLPipeline:
    pipeline_config = dict(
        core_name=core_name,
        core_config=core_config,
        input_dim=input_dim,
        output_dim=output_dim,
        loss_name=loss_name,
        loss_config=loss_config,
        only_categorical=only_categorical,
        encoder_config=encoder_config,
        encoding_settings=encoding_settings,
        state_config=state_config,
        num_epoch=num_epoch,
        max_epoch=max_epoch,
        fixed_epoch=fixed_epoch,
        fixed_steps=fixed_steps,
        log_steps=log_steps,
        valid_portion=valid_portion,
        amp=amp,
        clip_norm=clip_norm,
        cudnn_benchmark=cudnn_benchmark,
        metric_names=metric_names,
        metric_configs=metric_configs,
        metric_weights=metric_weights,
        use_losses_as_metrics=use_losses_as_metrics,
        loss_metrics_weights=loss_metrics_weights,
        recompute_train_losses_in_eval=recompute_train_losses_in_eval,
        monitor_names=monitor_names,
        monitor_configs=monitor_configs,
        callback_names=callback_names,
        callback_configs=callback_configs,
        lr=lr,
        optimizer_name=optimizer_name,
        scheduler_name=scheduler_name,
        optimizer_config=optimizer_config,
        scheduler_config=scheduler_config,
        optimizer_settings=optimizer_settings,
        use_zero=use_zero,
        workplace=workplace,
        finetune_config=finetune_config,
        tqdm_settings=tqdm_settings,
        pre_process_batch=pre_process_batch,
    )
    if debug:
        inject_debug(pipeline_config)
    fit_kwargs = dict(sample_weights=sample_weights, cuda=cuda)
    m_base = MLCarefreePipeline if carefree else MLPipeline
    data = _make_ml_data(
        x_train,
        y_train,
        x_valid,
        y_valid,
        train_others,
        valid_others,
        carefree,
        is_classification,
        data_config,
        cf_data_config,
    )
    return m_base(**pipeline_config).fit(data, **fit_kwargs)  # type: ignore


def repeat_ml(
    x_train: Union[data_type, IMLData],
    y_train: data_type = None,
    x_valid: data_type = None,
    y_valid: data_type = None,
    *,
    num_jobs: int = 1,
    num_repeat: int = 5,
    train_others: Optional[np_dict_type] = None,
    valid_others: Optional[np_dict_type] = None,
    # repeat
    workplace: str = "_repeat",
    models: Union[str, List[str]] = "fcnn",
    model_configs: Optional[Dict[str, Dict[str, Any]]] = None,
    predict_config: Optional[Dict[str, Any]] = None,
    sequential: Optional[bool] = None,
    return_patterns: bool = True,
    compress: bool = True,
    use_tqdm: bool = True,
    cuda: Optional[Union[int, str]] = None,
    available_cuda_list: Optional[List[int]] = None,
    resource_config: Optional[Dict[str, Any]] = None,
    task_meta_kwargs: Optional[Dict[str, Any]] = None,
    to_original_device: bool = False,
    is_fix: bool = False,
    # data
    carefree: bool = False,
    is_classification: Optional[bool] = None,
    data_config: Optional[Dict[str, Any]] = None,
    cf_data_config: Optional[Dict[str, Any]] = None,
    # pipeline
    input_dim: Optional[int] = None,
    output_dim: Optional[int] = None,
    loss_name: str = "auto",
    loss_config: Optional[Dict[str, Any]] = None,
    # encoder
    only_categorical: bool = False,
    encoder_config: Optional[Dict[str, Any]] = None,
    encoding_settings: Optional[Dict[int, Dict[str, Any]]] = None,
    # trainer
    state_config: Optional[Dict[str, Any]] = None,
    num_epoch: int = 40,
    max_epoch: int = 1000,
    fixed_epoch: Optional[int] = None,
    fixed_steps: Optional[int] = None,
    log_steps: Optional[int] = None,
    valid_portion: float = 1.0,
    amp: bool = False,
    clip_norm: float = 0.0,
    cudnn_benchmark: bool = False,
    metric_names: Optional[Union[str, List[str]]] = None,
    metric_configs: configs_type = None,
    metric_weights: Optional[Dict[str, float]] = None,
    use_losses_as_metrics: Optional[bool] = None,
    loss_metrics_weights: Optional[Dict[str, float]] = None,
    recompute_train_losses_in_eval: bool = True,
    monitor_names: Optional[Union[str, List[str]]] = None,
    monitor_configs: Optional[Dict[str, Any]] = None,
    callback_names: Optional[Union[str, List[str]]] = None,
    callback_configs: Optional[Dict[str, Any]] = None,
    lr: Optional[float] = None,
    optimizer_name: Optional[str] = None,
    scheduler_name: Optional[str] = None,
    optimizer_config: Optional[Dict[str, Any]] = None,
    scheduler_config: Optional[Dict[str, Any]] = None,
    optimizer_settings: Optional[Dict[str, Dict[str, Any]]] = None,
    finetune_config: Optional[Dict[str, Any]] = None,
    tqdm_settings: Optional[Dict[str, Any]] = None,
    # misc
    sample_weights: sample_weights_type = None,
    pre_process_batch: bool = True,
    debug: bool = False,
) -> RepeatResult:
    pipeline_config = dict(
        input_dim=input_dim,
        output_dim=output_dim,
        loss_name=loss_name,
        loss_config=loss_config,
        only_categorical=only_categorical,
        encoder_config=encoder_config,
        encoding_settings=encoding_settings,
        state_config=state_config,
        num_epoch=num_epoch,
        max_epoch=max_epoch,
        fixed_epoch=fixed_epoch,
        fixed_steps=fixed_steps,
        log_steps=log_steps,
        valid_portion=valid_portion,
        amp=amp,
        clip_norm=clip_norm,
        cudnn_benchmark=cudnn_benchmark,
        metric_names=metric_names,
        metric_configs=metric_configs,
        metric_weights=metric_weights,
        use_losses_as_metrics=use_losses_as_metrics,
        loss_metrics_weights=loss_metrics_weights,
        recompute_train_losses_in_eval=recompute_train_losses_in_eval,
        monitor_names=monitor_names,
        monitor_configs=monitor_configs,
        callback_names=callback_names,
        callback_configs=callback_configs,
        lr=lr,
        optimizer_name=optimizer_name,
        scheduler_name=scheduler_name,
        optimizer_config=optimizer_config,
        scheduler_config=scheduler_config,
        optimizer_settings=optimizer_settings,
        finetune_config=finetune_config,
        tqdm_settings=tqdm_settings,
        pre_process_batch=pre_process_batch,
    )
    if debug:
        inject_debug(pipeline_config)
    return repeat_with(
        _make_ml_data(
            x_train,
            y_train,
            x_valid,
            y_valid,
            train_others,
            valid_others,
            carefree,
            is_classification,
            data_config,
            cf_data_config,
        ),
        carefree=carefree,
        workplace=workplace,
        models=models,
        model_configs=model_configs,
        predict_config=predict_config,
        sample_weights=sample_weights,
        sequential=sequential,
        num_jobs=num_jobs,
        num_repeat=num_repeat,
        return_patterns=return_patterns,
        compress=compress,
        use_tqdm=use_tqdm,
        cuda=cuda,
        available_cuda_list=available_cuda_list,
        resource_config=resource_config,
        task_meta_kwargs=task_meta_kwargs,
        to_original_device=to_original_device,
        is_fix=is_fix,
        **pipeline_config,
    )


def supported_ml_models() -> List[str]:
    return sorted(ml_core_dict)


# cv


def fit_cv(
    data: CVDataModule,
    model_name: str,
    model_config: Dict[str, Any],
    *,
    loss_name: Optional[str] = None,
    loss_config: Optional[Dict[str, Any]] = None,
    # trainer
    state_config: Optional[Dict[str, Any]] = None,
    num_epoch: int = 40,
    max_epoch: int = 1000,
    fixed_epoch: Optional[int] = None,
    fixed_steps: Optional[int] = None,
    log_steps: Optional[int] = None,
    valid_portion: float = 1.0,
    amp: bool = False,
    clip_norm: float = 0.0,
    cudnn_benchmark: bool = False,
    metric_names: Optional[Union[str, List[str]]] = None,
    metric_configs: configs_type = None,
    metric_weights: Optional[Dict[str, float]] = None,
    use_losses_as_metrics: Optional[bool] = None,
    loss_metrics_weights: Optional[Dict[str, float]] = None,
    recompute_train_losses_in_eval: bool = True,
    monitor_names: Optional[Union[str, List[str]]] = None,
    monitor_configs: Optional[Dict[str, Any]] = None,
    callback_names: Optional[Union[str, List[str]]] = None,
    callback_configs: Optional[Dict[str, Any]] = None,
    lr: Optional[float] = None,
    optimizer_name: Optional[str] = None,
    scheduler_name: Optional[str] = None,
    optimizer_config: Optional[Dict[str, Any]] = None,
    scheduler_config: Optional[Dict[str, Any]] = None,
    optimizer_settings: Optional[Dict[str, Dict[str, Any]]] = None,
    use_zero: bool = False,
    workplace: str = "_logs",
    finetune_config: Optional[Dict[str, Any]] = None,
    tqdm_settings: Optional[Dict[str, Any]] = None,
    # misc
    debug: bool = False,
    # fit
    sample_weights: sample_weights_type = None,
    cuda: Optional[Union[int, str]] = None,
) -> "CVPipeline":
    pipeline_config = dict(
        model_name=model_name,
        model_config=model_config,
        loss_name=loss_name,
        loss_config=loss_config,
        state_config=state_config,
        num_epoch=num_epoch,
        max_epoch=max_epoch,
        fixed_epoch=fixed_epoch,
        fixed_steps=fixed_steps,
        log_steps=log_steps,
        valid_portion=valid_portion,
        amp=amp,
        clip_norm=clip_norm,
        cudnn_benchmark=cudnn_benchmark,
        metric_names=metric_names,
        metric_configs=metric_configs,
        metric_weights=metric_weights,
        use_losses_as_metrics=use_losses_as_metrics,
        loss_metrics_weights=loss_metrics_weights,
        recompute_train_losses_in_eval=recompute_train_losses_in_eval,
        monitor_names=monitor_names,
        monitor_configs=monitor_configs,
        callback_names=callback_names,
        callback_configs=callback_configs,
        lr=lr,
        optimizer_name=optimizer_name,
        scheduler_name=scheduler_name,
        optimizer_config=optimizer_config,
        scheduler_config=scheduler_config,
        optimizer_settings=optimizer_settings,
        use_zero=use_zero,
        workplace=workplace,
        finetune_config=finetune_config,
        tqdm_settings=tqdm_settings,
    )
    if debug:
        inject_debug(pipeline_config)
    fit_kwargs = dict(sample_weights=sample_weights, cuda=cuda)
    return CVPipeline(**pipeline_config).fit(data, **fit_kwargs)  # type: ignore


# clf


def _clf(
    model: str,
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]],
    pretrained_name: Optional[str],
    img_size: Optional[int],
    return_model: bool = False,
    **kwargs: Any,
) -> Any:
    if img_size is not None:
        kwargs["img_size"] = img_size
    kwargs["num_classes"] = num_classes
    if pretrained_name is not None:
        model_config = kwargs.setdefault("model_config", {})
        model_config["encoder1d_pretrained_name"] = pretrained_name
    model = f"clf/{model}"
    fn = DLZoo.load_model if return_model else DLZoo.load_pipeline
    if aux_num_classes is None:
        return fn(model, **kwargs)  # type: ignore
    config = DLZoo(model, **kwargs).config
    aux_labels = sorted(aux_num_classes)
    loss_name = config["loss_name"]
    config["loss_name"] = f"{loss_name}:aux:{','.join(aux_labels)}"
    metric_names = config.setdefault("metric_names", [])
    metric_configs = config.setdefault("metric_configs", {})
    if isinstance(metric_configs, dict):
        metric_configs = [metric_configs.get(name, {}) for name in metric_names]
    for label in aux_labels:
        metric_names.append("aux")
        metric_configs.append({"key": label, "base": "acc"})
    config["metric_configs"] = metric_configs
    model_config = config.setdefault("model_config", {})
    model_config["aux_num_classes"] = aux_num_classes
    return fn(model, **config)  # type: ignore


def cct(
    img_size: int,
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    **kwargs: Any,
) -> DLPipeline:
    return _clf("cct", num_classes, aux_num_classes, None, img_size, **kwargs)


def cct_model(
    img_size: int,
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    **kwargs: Any,
) -> VanillaClassifier:
    return _clf("cct", num_classes, aux_num_classes, None, img_size, True, **kwargs)


def cct_lite(
    img_size: int,
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    **kwargs: Any,
) -> DLPipeline:
    return _clf("cct.lite", num_classes, aux_num_classes, None, img_size, **kwargs)


def cct_lite_model(
    img_size: int,
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    **kwargs: Any,
) -> VanillaClassifier:
    return _clf(
        "cct.lite",
        num_classes,
        aux_num_classes,
        None,
        img_size,
        True,
        **kwargs,
    )


def cct_large(
    img_size: int,
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    **kwargs: Any,
) -> DLPipeline:
    return _clf("cct.large", num_classes, aux_num_classes, None, img_size, **kwargs)


def cct_large_model(
    img_size: int,
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    **kwargs: Any,
) -> VanillaClassifier:
    return _clf(
        "cct.large",
        num_classes,
        aux_num_classes,
        None,
        img_size,
        True,
        **kwargs,
    )


def cct_large_224(
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    *,
    pretrained: bool = True,
    **kwargs: Any,
) -> DLPipeline:
    pretrained_name = "cct_large_224" if pretrained else None
    return _clf(
        "cct.large_224",
        num_classes,
        aux_num_classes,
        pretrained_name,
        None,
        **kwargs,
    )


def cct_large_224_model(
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    *,
    pretrained: bool = True,
    **kwargs: Any,
) -> VanillaClassifier:
    pretrained_name = "cct_large_224" if pretrained else None
    return _clf(
        "cct.large_224",
        num_classes,
        aux_num_classes,
        pretrained_name,
        None,
        True,
        **kwargs,
    )


def cct_large_384(
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    *,
    pretrained: bool = True,
    **kwargs: Any,
) -> DLPipeline:
    pretrained_name = "cct_large_384" if pretrained else None
    return _clf(
        "cct.large_384",
        num_classes,
        aux_num_classes,
        pretrained_name,
        None,
        **kwargs,
    )


def cct_large_384_model(
    num_classes: int,
    aux_num_classes: Optional[Dict[str, int]] = None,
    *,
    pretrained: bool = True,
    **kwargs: Any,
) -> VanillaClassifier:
    pretrained_name = "cct_large_384" if pretrained else None
    return _clf(
        "cct.large_384",
        num_classes,
        aux_num_classes,
        pretrained_name,
        None,
        True,
        **kwargs,
    )


def resnet18(num_classes: int, pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    kwargs["num_classes"] = num_classes
    model_config = kwargs.setdefault("model_config", {})
    encoder1d_config = model_config.setdefault("encoder1d_config", {})
    encoder1d_config["pretrained"] = pretrained
    return DLZoo.load_pipeline("clf/resnet18", **kwargs)


def resnet18_model(
    num_classes: int,
    *,
    pretrained: bool = False,
    **kwargs: Any,
) -> VanillaClassifier:
    kwargs["num_classes"] = num_classes
    model_config = kwargs.setdefault("model_config", {})
    encoder1d_config = model_config.setdefault("encoder1d_config", {})
    encoder1d_config["pretrained"] = pretrained
    return DLZoo.load_model("clf/resnet18", **kwargs)


def resnet18_gray(num_classes: int, **kwargs: Any) -> DLPipeline:
    kwargs["num_classes"] = num_classes
    return DLZoo.load_pipeline("clf/resnet18.gray", **kwargs)


def resnet50(num_classes: int, pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    kwargs["num_classes"] = num_classes
    model_config = kwargs.setdefault("model_config", {})
    encoder1d_config = model_config.setdefault("encoder1d_config", {})
    encoder1d_config["pretrained"] = pretrained
    return DLZoo.load_pipeline("clf/resnet50", **kwargs)


def resnet50_model(
    num_classes: int,
    *,
    pretrained: bool = False,
    **kwargs: Any,
) -> VanillaClassifier:
    kwargs["num_classes"] = num_classes
    model_config = kwargs.setdefault("model_config", {})
    encoder1d_config = model_config.setdefault("encoder1d_config", {})
    encoder1d_config["pretrained"] = pretrained
    return DLZoo.load_model("clf/resnet50", **kwargs)


def resnet50_gray(num_classes: int, **kwargs: Any) -> DLPipeline:
    kwargs["num_classes"] = num_classes
    return DLZoo.load_pipeline("clf/resnet50.gray", **kwargs)


def resnet101(num_classes: int, pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    kwargs["num_classes"] = num_classes
    model_config = kwargs.setdefault("model_config", {})
    encoder1d_config = model_config.setdefault("encoder1d_config", {})
    encoder1d_config["pretrained"] = pretrained
    return DLZoo.load_pipeline("clf/resnet101", **kwargs)


def resnet101_model(
    num_classes: int,
    *,
    pretrained: bool = False,
    **kwargs: Any,
) -> VanillaClassifier:
    kwargs["num_classes"] = num_classes
    model_config = kwargs.setdefault("model_config", {})
    encoder1d_config = model_config.setdefault("encoder1d_config", {})
    encoder1d_config["pretrained"] = pretrained
    return DLZoo.load_model("clf/resnet101", **kwargs)


# gan


def vanilla_gan(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("gan/vanilla", **kwargs)


def vanilla_gan_gray(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("gan/vanilla.gray", **kwargs)


def siren_gan(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("gan/siren", **kwargs)


def siren_gan_gray(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("gan/siren.gray", **kwargs)


# generator


def pixel_cnn(num_classes: int, **kwargs: Any) -> DLPipeline:
    kwargs["num_classes"] = num_classes
    return DLZoo.load_pipeline("generator/pixel_cnn", **kwargs)


# multimodal


def clip(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    return DLZoo.load_pipeline("multimodal/clip", pretrained=pretrained, **kwargs)


def clip_model(pretrained: bool = True, **kwargs: Any) -> CLIP:
    return DLZoo.load_model("multimodal/clip", pretrained=pretrained, **kwargs)


def clip_large(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    return DLZoo.load_pipeline("multimodal/clip.large", pretrained=pretrained, **kwargs)


def clip_large_model(pretrained: bool = True, **kwargs: Any) -> CLIP:
    return DLZoo.load_model("multimodal/clip.large", pretrained=pretrained, **kwargs)


def chinese_clip(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    model = "multimodal/clip.chinese"
    return DLZoo.load_pipeline(model, pretrained=pretrained, **kwargs)


def chinese_clip_model(pretrained: bool = True, **kwargs: Any) -> CLIP:
    model = "multimodal/clip.chinese"
    return DLZoo.load_model(model, pretrained=pretrained, **kwargs)


def clip_vqgan_aligner(**kwargs: Any) -> DLPipeline:
    return DLZoo.load_pipeline("multimodal/clip_vqgan_aligner", **kwargs)


# segmentor


def aim() -> DLPipeline:
    return DLZoo.load_pipeline("segmentor/aim")


def aim_model() -> IDLModel:
    return DLZoo.load_model("segmentor/aim")


def u2net(pretrained: bool = False, **kwargs: Any) -> DLPipeline:
    return DLZoo.load_pipeline("segmentor/u2net", pretrained=pretrained, **kwargs)


def u2net_model(pretrained: bool = False, **kwargs: Any) -> IDLModel:
    return DLZoo.load_model("segmentor/u2net", pretrained=pretrained, **kwargs)


def u2net_lite(pretrained: bool = False, **kwargs: Any) -> DLPipeline:
    return DLZoo.load_pipeline("segmentor/u2net.lite", pretrained=pretrained, **kwargs)


def u2net_lite_model(pretrained: bool = False, **kwargs: Any) -> IDLModel:
    return DLZoo.load_model("segmentor/u2net.lite", pretrained=pretrained, **kwargs)


def u2net_finetune(ckpt: Optional[str] = None, **kwargs: Any) -> DLPipeline:
    if ckpt is None:
        ckpt = download_model("u2net")
    kwargs["pretrained_ckpt"] = ckpt
    return DLZoo.load_pipeline("segmentor/u2net.finetune", **kwargs)


def u2net_lite_finetune(ckpt: Optional[str] = None, **kwargs: Any) -> DLPipeline:
    if ckpt is None:
        ckpt = download_model("u2net.lite")
    kwargs["pretrained_ckpt"] = ckpt
    return DLZoo.load_pipeline("segmentor/u2net.finetune_lite", **kwargs)


def u2net_refine(lv1_model_ckpt_path: str, **kwargs: Any) -> DLPipeline:
    kwargs["lv1_model_ckpt_path"] = lv1_model_ckpt_path
    return DLZoo.load_pipeline("segmentor/u2net.refine", **kwargs)


def u2net_lite_refine(lv1_model_ckpt_path: str, **kwargs: Any) -> DLPipeline:
    kwargs["lv1_model_ckpt_path"] = lv1_model_ckpt_path
    return DLZoo.load_pipeline("segmentor/u2net.refine_lite", **kwargs)


# ssl


def dino(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("ssl/dino", **kwargs)


def dino_gray(img_size: int, **kwargs: Any) -> DLPipeline:
    model_config = kwargs.setdefault("model_config", {})
    encoder1d_config = model_config.setdefault("encoder1d_config", {})
    encoder1d_config["in_channels"] = 1
    return dino(img_size, **kwargs)


# style transfer


def adain(pretrained: bool = False, **kwargs: Any) -> DLPipeline:
    kwargs["pretrained"] = pretrained
    return DLZoo.load_pipeline("style_transfer/adain", **kwargs)


# ae


def ae_kl_f4(size: int = 256, pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_kl_f4` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return DLZoo.load_pipeline("ae/kl.f4", **kwargs)


def ae_kl_f8(size: int = 256, pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_kl_f8` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return DLZoo.load_pipeline("ae/kl.f8", **kwargs)


def ae_kl_f16(size: int = 256, pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_kl_f16` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return DLZoo.load_pipeline("ae/kl.f16", **kwargs)


def ae_vq_f4(size: int = 256, pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_vq_f4` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return DLZoo.load_pipeline("ae/vq.f4", **kwargs)


def ae_vq_f4_no_attn(
    size: int = 256,
    pretrained: bool = True,
    **kwargs: Any,
) -> DLPipeline:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_vq_f4_no_attn` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return DLZoo.load_pipeline("ae/vq.f4_no_attn", **kwargs)


def ae_vq_f8(size: int = 256, pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    if pretrained and size != 256:
        raise ValueError("pretrained `ae_vq_f8` should have `size`=256")
    kwargs["img_size"] = size
    kwargs["pretrained"] = pretrained
    return DLZoo.load_pipeline("ae/vq.f8", **kwargs)


# vae


def vanilla_vae(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("vae/vanilla", **kwargs)


def vanilla_vae_gray(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("vae/vanilla.gray", **kwargs)


def vanilla_vae2d(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("vae/vanilla.2d", **kwargs)


def vanilla_vae2d_gray(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("vae/vanilla.2d_gray", **kwargs)


def style_vae(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("vae/style", **kwargs)


def style_vae_gray(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("vae/style.gray", **kwargs)


def siren_vae(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("vae/siren", **kwargs)


def siren_vae_gray(img_size: int, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("vae/siren.gray", **kwargs)


def _vq_vae(
    model: str,
    img_size: int,
    num_classes: Optional[int] = None,
    **kwargs: Any,
) -> DLPipeline:
    kwargs["img_size"] = img_size
    if num_classes is not None:
        model_config = kwargs.setdefault("model_config", {})
        num_classes = model_config.setdefault("num_classes", num_classes)
        callback_names = kwargs.get("callback_names")
        if callback_names is None or "vq_vae" in callback_names:
            callback_configs = kwargs.setdefault("callback_configs", {})
            vq_vae_callback_configs = callback_configs.setdefault("vq_vae", {})
            vq_vae_callback_configs.setdefault("num_classes", num_classes)
    return DLZoo.load_pipeline(f"vae/{model}", **kwargs)


def vq_vae(
    img_size: int,
    *,
    num_classes: Optional[int] = None,
    **kwargs: Any,
) -> DLPipeline:
    return _vq_vae("vq", img_size, num_classes, **kwargs)


def vq_vae_gray(
    img_size: int,
    *,
    num_classes: Optional[int] = None,
    **kwargs: Any,
) -> DLPipeline:
    return _vq_vae("vq.gray", img_size, num_classes, **kwargs)


def vq_vae_lite(
    img_size: int,
    *,
    num_classes: Optional[int] = None,
    **kwargs: Any,
) -> DLPipeline:
    return _vq_vae("vq.lite", img_size, num_classes, **kwargs)


def vq_vae_gray_lite(
    img_size: int,
    *,
    num_classes: Optional[int] = None,
    **kwargs: Any,
) -> DLPipeline:
    return _vq_vae("vq.gray_lite", img_size, num_classes, **kwargs)


# diffusion


def ddpm(img_size: int = 256, **kwargs: Any) -> DLPipeline:
    kwargs["img_size"] = img_size
    return DLZoo.load_pipeline("diffusion/ddpm", **kwargs)


def ldm(
    latent_size: int = 32,
    latent_in_channels: int = 4,
    latent_out_channels: int = 4,
    **kwargs: Any,
) -> DLPipeline:
    return _ldm(latent_size, latent_in_channels, latent_out_channels, **kwargs)


def ldm_vq(
    latent_size: int = 64,
    latent_in_channels: int = 3,
    latent_out_channels: int = 3,
    **kwargs: Any,
) -> DLPipeline:
    return _ldm_vq(latent_size, latent_in_channels, latent_out_channels, **kwargs)


def ldm_sd(pretrained: bool = True) -> DLPipeline:
    return _ldm_sd(pretrained)


def ldm_sd_anime(pretrained: bool = True) -> DLPipeline:
    return _ldm_sd_anime(pretrained)


def ldm_celeba_hq(pretrained: bool = True) -> DLPipeline:
    return _ldm_celeba_hq(pretrained)


def ldm_inpainting(pretrained: bool = True) -> DLPipeline:
    return _ldm_inpainting(pretrained)


def ldm_sr(pretrained: bool = True) -> DLPipeline:
    return _ldm_sr(pretrained)


def ldm_semantic(pretrained: bool = True) -> DLPipeline:
    return _ldm_semantic(pretrained)


# sr


def esr(pretrained: bool = True) -> DLPipeline:
    return _esr(pretrained)


def esr_anime(pretrained: bool = True) -> DLPipeline:
    return _esr_anime(pretrained)


# nlp


def hugging_face(model: str) -> DLPipeline:
    return DLZoo.load_pipeline("hugging_face/general", model_config={"model": model})


def hugging_face_model(model: str) -> HuggingFaceModel:
    return DLZoo.load_model("hugging_face/general", model_config={"model": model})


def simbert_model() -> SimBERT:
    return DLZoo.load_model("hugging_face/simbert")


def opus_model(src: str, tgt: str) -> OPUSBase:
    return DLZoo.load_model("hugging_face/opus", model_config={"src": src, "tgt": tgt})


def opus_zh_en_model() -> OPUS_ZH_EN:
    return opus_model("zh", "en")



import os

import numpy as np

from typing import Any
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from cftool.array import to_numpy
from cftool.array import to_torch
from cftool.types import tensor_dict_type

from ..pipeline import MLPipeline
from ....constants import PREDICTIONS_KEY
from ....misc.toolkit import eval_context
from ....models.ml import DDR

try:
    import matplotlib.pyplot as plt
    from cfml.misc.toolkit import show_or_save

    Figure = plt.Figure
except:
    plt = show_or_save = Figure = None


class DDRPredictor:
    def __init__(self, ddr: DDR):
        self.m = ddr
        self.device = ddr.device

    def _fetch(self, x: np.ndarray, **kwargs: Any) -> tensor_dict_type:
        net = to_torch(x).to(self.device)
        with eval_context(self.m, use_grad=True):
            return self.m(net, **kwargs)

    def median(self, x: np.ndarray) -> np.ndarray:
        results = self._fetch(x, get_quantiles=False)
        return to_numpy(results[PREDICTIONS_KEY])

    def quantile(self, x: np.ndarray, tau: Union[float, List[float]]) -> np.ndarray:
        results = self._fetch(x, tau=tau, get_cdf=False)
        return to_numpy(results["quantiles"])

    def cdf_pdf(
        self,
        x: np.ndarray,
        y: Union[float, List[float]],
    ) -> Tuple[np.ndarray, np.ndarray]:
        if isinstance(y, float):
            y = [y]
        y_anchor = np.array(y, np.float32).reshape([1, -1])
        y_anchor = np.repeat(y_anchor, len(x), axis=0)
        results = self._fetch(x, y_anchor=y_anchor, get_quantiles=False)
        return to_numpy(results["cdf"]), to_numpy(results["pdf"])

    @classmethod
    def from_pipeline(cls, m: MLPipeline) -> "DDRPredictor":
        return cls(m.model.core.core)


class DDRVisualizer:
    def __init__(
        self,
        ddr: DDR,
        dpi: int = 200,
        figsize: Tuple[int, int] = (8, 6),
    ):
        if show_or_save is None:
            raise ValueError("`carefree-ml` is needed for `DDRVisualizer`")
        self.m = ddr
        self.dpi = dpi
        self.figsize = figsize
        self.predictor = DDRPredictor(ddr)

    def _prepare_base_figure(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_base: np.ndarray,
        mean: Optional[np.ndarray],
        median: Optional[np.ndarray],
        indices: np.ndarray,
        title: str,
    ) -> Figure:  # type: ignore
        figure = plt.figure(figsize=self.figsize, dpi=self.dpi)
        plt.title(title)
        plt.scatter(x[indices], y[indices], color="gray", s=15)
        if mean is not None:
            plt.plot(x_base.ravel(), mean.ravel(), label="mean")
        if median is not None:
            plt.plot(x_base.ravel(), median.ravel(), label="median")
        return figure

    @staticmethod
    def _render_figure(
        x_min: np.ndarray,
        x_max: np.ndarray,
        y_min: float,
        y_max: float,
        y_padding: float,
    ) -> None:
        plt.xlim(x_min, x_max)
        plt.ylim(y_min - 0.5 * y_padding, y_max + 0.5 * y_padding)
        plt.legend()

    def visualize(
        self,
        x: np.ndarray,
        y: np.ndarray,
        export_path: Optional[str],
        *,
        ratios: Union[float, List[float]],
        **kwargs: Any,
    ) -> None:
        if isinstance(ratios, float):
            ratios = [ratios]

        x_min, x_max = np.min(x), np.max(x)
        y_min, y_max = np.min(y), np.max(y)
        padding, dense = kwargs.get("padding", 1.0), kwargs.get("dense", 400)
        indices = np.random.permutation(len(x))[:dense]
        x_padding = max(abs(x_min), abs(x_max)) * padding
        y_padding = max(abs(y_min), abs(y_max)) * padding
        x_min -= x_padding
        x_max += x_padding
        x_base = np.linspace(x_min, x_max, dense)[..., None]
        mean = None
        median = self.predictor.median(x_base)
        assert isinstance(median, np.ndarray)
        fig = self._prepare_base_figure(x, y, x_base, mean, median, indices, "")
        render_args = x_min, x_max, y_min, y_max, y_padding
        # quantile curves
        if self.m.predict_quantiles:
            quantile_curves = self.predictor.quantile(x_base, ratios)
            if quantile_curves.shape[-1] != 1:
                raise ValueError("`out_dim` > 1 is not supported")
            quantile_curves = quantile_curves[..., 0]
            for q, quantile_curve in zip(ratios, quantile_curves.T):
                plt.plot(x_base.ravel(), quantile_curve, label=f"quantile {q:4.2f}")
            DDRVisualizer._render_figure(*render_args)
        # cdf curves
        if self.m.predict_cdf:
            anchors = [ratio * (y_max - y_min) + y_min for ratio in ratios]
            cdfs, pdfs = self.predictor.cdf_pdf(x_base, anchors)
            if cdfs.shape[-1] != 1 or pdfs.shape[-1] != 1:
                raise ValueError("`out_dim` > 1 is not supported")
            cdfs, pdfs = cdfs[..., 0], pdfs[..., 0]
            for ratio, anchor, cdf, pdf in zip(ratios, anchors, cdfs.T, pdfs.T):
                anchor_line = np.full(len(x_base), anchor)
                plt.plot(x_base.ravel(), cdf, label=f"cdf {ratio:4.2f}")
                pdf = pdf / (pdf.max() + 1.0e-8) * y_max
                plt.plot(x_base.ravel(), pdf, label=f"pdf {ratio:4.2f}")
                plt.plot(
                    x_base.ravel(),
                    anchor_line,
                    label=f"anchor {ratio:4.2f}",
                    color="gray",
                )
            DDRVisualizer._render_figure(*render_args)
        show_or_save(export_path, fig)

    def visualize_multiple(
        self,
        f: Callable,
        x: np.ndarray,
        y: np.ndarray,
        export_folder: str,
        *,
        num_base: int = 1000,
        num_repeat: int = 10000,
        ratios: Union[float, List[float]],
    ) -> None:
        if isinstance(ratios, float):
            ratios = [ratios]

        x_min, x_max = x.min(), x.max()
        x_diff = x_max - x_min
        x_base = np.linspace(x_min - 0.1 * x_diff, x_max + 0.1 * x_diff, num_base)
        x_base = x_base[..., None]
        x_matrix = np.repeat(x_base, num_repeat, axis=1)
        y_matrix = f(x_matrix)
        y_min, y_max = y.min(), y.max()
        y_diff = y_max - y_min

        def _plot(
            prefix: str,
            num: float,
            y_true: Optional[np.ndarray],
            predictions: np.ndarray,
            anchor_line_: Optional[np.ndarray],
        ) -> None:
            plt.figure(figsize=self.figsize, dpi=self.dpi)
            plt.title(f"{prefix} {num:6.4f}")
            plt.scatter(x[:200], y[:200], color="gray", s=15)
            if y_true is not None:
                plt.plot(x_base, y_true, label="target")
            plt.plot(x_base, predictions, label=f"ddr_prediction")
            if anchor_line_ is not None:
                plt.plot(x_base.ravel(), anchor_line_, color="gray")
            plt.legend()
            show_or_save(os.path.join(export_folder, f"{prefix}_{num:4.2f}.png"))

        if self.m.predict_quantiles:
            yq_predictions = self.predictor.quantile(x_base, ratios)
            yq_predictions = yq_predictions[..., 0]
            for q, yq_pred in zip(ratios, yq_predictions.T):
                yq = np.percentile(y_matrix, int(100 * q), axis=1)
                _plot("quantile", q, yq, yq_pred, None)

        if self.m.predict_cdf:
            anchors = [ratio * (y_max - y_min) + y_min for ratio in ratios]
            for anchor in anchors:
                assert isinstance(anchor, float)
                anchor_line = np.full(len(x_base), anchor)
                yd = np.mean(y_matrix <= anchor, axis=1) * y_diff + y_min
                cdf, pdf = self.predictor.cdf_pdf(x_base, anchor)
                cdf, pdf = cdf[..., 0], pdf[..., 0]
                cdf = cdf * y_diff + y_min
                _plot("cdf", anchor, yd, cdf, anchor_line)
                pdf = pdf / (pdf.max() + 1.0e-8) * y_max
                _plot("pdf", anchor, None, pdf, anchor_line)

    @classmethod
    def from_pipeline(
        cls,
        m: MLPipeline,
        dpi: int = 200,
        figsize: Tuple[int, int] = (8, 6),
    ) -> "DDRVisualizer":
        return cls(m.model.core.core, dpi, figsize)


__all__ = [
    "DDRPredictor",
    "DDRVisualizer",
]



from setuptools import setup
from setuptools import find_packages


VERSION = "0.3.3.2"
DESCRIPTION = "Deep Learning with PyTorch made easy"
with open("README.md", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="carefree-learn",
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    install_requires=["carefree-toolkit>=0.2.17"],
    extras_require={
        "onnx": [
            "onnx",
            "onnx-simplifier>=0.4.1",
            "onnxruntime",
        ],
        "cv": [
            "ftfy",
            "regex",
            "lmdb",
            "albumentations",
            "onnx",
            "onnx-simplifier>=0.4.1",
            "onnxruntime",
            "carefree-cv>=0.1.0",
        ],
        "full": [
            "faiss-cpu",
            "protobuf==3.19.4",
            "ortools>=9.3.0",
            "sacremoses",
            "sentencepiece",
            "transformers",
            "ftfy",
            "regex",
            "lmdb",
            "albumentations",
            "mlflow",
            "onnx",
            "onnx-simplifier>=0.4.1",
            "onnxruntime",
            "plotly",
            "carefree-data>=0.2.9",
            "carefree-ml>=0.1.3",
            "carefree-cv>=0.1.0",
        ],
    },
    author="carefree0910",
    author_email="syameimaru.saki@gmail.com",
    url="https://github.com/carefree0910/carefree-learn",
    download_url=f"https://github.com/carefree0910/carefree-learn/archive/v{VERSION}.tar.gz",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="python machine-learning deep-learning solution PyTorch",
)



from .misc import *
from .losses import *
from .models import *
from .modules import *
from .trainer import *
from .protocol import *
from .constants import *
from .data import *
from .pipeline import *
from .zoo import *

from .api import cv
from .api import ml
from .api import nlp
from .api import multimodal
from .api import api
from . import dist
from . import scripts

from pkg_resources import get_distribution

pkg = get_distribution("carefree-learn")
__version__ = pkg.version



import torch

import numpy as np
import torch.nn as nn

from abc import abstractmethod
from PIL import Image
from typing import Any
from typing import Dict
from typing import Type
from typing import Tuple
from typing import Union
from typing import TypeVar
from typing import Optional
from typing import NamedTuple
from cftool.misc import WithRegister
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from torch.cuda.amp.autocast_mode import autocast

from ....pipeline import DLPipeline

try:
    import cv2
except:
    cv2 = None
try:
    from cfcv.misc.toolkit import to_rgb
except:
    to_rgb = None


def restrict_wh(w: int, h: int, max_wh: int) -> Tuple[int, int]:
    max_original_wh = max(w, h)
    if max_original_wh <= max_wh:
        return w, h
    wh_ratio = w / h
    if wh_ratio >= 1:
        return max_wh, round(max_wh / wh_ratio)
    return round(max_wh * wh_ratio), max_wh


def get_suitable_size(n: int, anchor: int) -> int:
    mod = n % anchor
    return n - mod + int(mod > 0.5 * anchor) * anchor


class ReadImageResponse(NamedTuple):
    image: np.ndarray
    alpha: Optional[np.ndarray]
    original_size: Tuple[int, int]


padding_modes: Dict[str, Type["Padding"]] = {}


class Padding(WithRegister):
    d = padding_modes

    @abstractmethod
    def pad(self, image: Image.Image, alpha: Image.Image, **kwargs: Any) -> Image.Image:
        pass


@Padding.register("cv2_ns")
class CV2NS(Padding):
    def pad(
        self,
        image: Image.Image,
        alpha: Image.Image,
        *,
        radius: int = 5,
        **kwargs: Any,
    ) -> Image.Image:
        if cv2 is None:
            raise ValueError("`cv2` is needed for `CV2NS`")
        img_arr = np.array(image.convert("RGB"))[..., ::-1]
        mask_arr = np.array(alpha)
        rs = cv2.inpaint(img_arr, 255 - mask_arr, radius, cv2.INPAINT_NS)
        return Image.fromarray(rs[..., ::-1])


@Padding.register("cv2_telea")
class CV2Telea(Padding):
    def pad(
        self,
        image: Image.Image,
        alpha: Image.Image,
        *,
        radius: int = 5,
        **kwargs: Any,
    ) -> Image.Image:
        if cv2 is None:
            raise ValueError("`cv2` is needed for `CV2Telea`")
        img_arr = np.array(image.convert("RGB"))[..., ::-1]
        mask_arr = np.array(alpha)
        rs = cv2.inpaint(img_arr, 255 - mask_arr, radius, cv2.INPAINT_TELEA)
        return Image.fromarray(rs[..., ::-1])


def read_image(
    image: Union[str, Image.Image],
    max_wh: int,
    *,
    anchor: int,
    to_gray: bool = False,
    to_mask: bool = False,
    resample: Any = Image.LANCZOS,
    normalize: bool = True,
    padding_mode: Optional[str] = None,
    padding_kwargs: Optional[Dict[str, Any]] = None,
) -> ReadImageResponse:
    if to_rgb is None:
        raise ValueError("`carefree-cv` is needed for `DiffusionAPI`")
    if isinstance(image, str):
        image = Image.open(image)
    alpha = None
    if image.mode == "RGBA":
        alpha = image.split()[3]
    if not to_mask and not to_gray:
        if alpha is None or padding_mode is None:
            image = to_rgb(image)
        else:
            padding = Padding.make(padding_mode, {})
            padding_kw = shallow_copy_dict(padding_kwargs or {})
            padding_kw.update(dict(image=image, alpha=alpha))
            image = safe_execute(padding.pad, padding_kw)
    else:
        if to_mask and to_gray:
            raise ValueError("`to_mask` & `to_gray` should not be True simultaneously")
        if to_mask and image.mode == "RGBA":
            image = alpha
        else:
            image = image.convert("L")
    original_w, original_h = image.size
    w, h = restrict_wh(original_w, original_h, max_wh)
    w, h = map(get_suitable_size, (w, h), (anchor, anchor))
    image = image.resize((w, h), resample=resample)
    image = np.array(image)
    if normalize:
        image = image.astype(np.float32) / 255.0
    if alpha is not None:
        alpha = (np.array(alpha).astype(np.float32) / 255.0)[None, None]
    if to_mask or to_gray:
        image = image[None, None]
    else:
        image = image[None].transpose(0, 3, 1, 2)
    return ReadImageResponse(image, alpha, (original_w, original_h))


T = TypeVar("T", bound="APIMixin")


class APIMixin:
    m: nn.Module
    device: torch.device
    use_amp: bool
    use_half: bool

    def __init__(
        self,
        m: nn.Module,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ):
        if use_amp and use_half:
            raise ValueError("`use_amp` & `use_half` should not be True simultaneously")
        self.m = m
        self.device = device
        self.use_amp = use_amp
        self.use_half = use_half

    @property
    def amp_context(self) -> autocast:
        return autocast(enabled=self.use_amp)

    @classmethod
    def from_pipeline(
        cls: Type[T],
        m: DLPipeline,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> T:
        if use_amp and use_half:
            raise ValueError("`use_amp` & `use_half` should not be True simultaneously")
        if use_half:
            m.model.half()
        if device is not None:
            m.model.to(device)
        return cls(m.model.core, m.model.device, use_amp=use_amp, use_half=use_half)



import torch

import torch.nn as nn

from torch import Tensor
from typing import List
from typing import Union
from typing import Optional

from ...protocols.cv import ImageTranslatorMixin
from ....misc.toolkit import interpolate
from ....modules.blocks import get_clones
from ....modules.blocks import get_conv_blocks
from ....modules.blocks import Conv2d
from ....modules.blocks import Interpolate
from ....misc.internal_.register import register_module


class ConvSeq(nn.Sequential):
    def __init__(self, in_channels: int = 3, out_channels: int = 3, dilation: int = 1):
        super().__init__(
            *get_conv_blocks(
                in_channels,
                out_channels,
                3,
                1,
                norm_type="batch",
                activation=nn.ReLU(inplace=True),
                padding=dilation,
                dilation=dilation,
            )
        )


class UNetBase(nn.Module):
    in_block: nn.Module
    down_blocks: nn.ModuleList
    last_down: nn.Module
    up_blocks: nn.ModuleList
    return_up_nets: bool = False

    def forward(
        self,
        net: Tensor,
        *,
        determinate: bool = False,
    ) -> Union[Tensor, List[Tensor]]:
        net = in_net = self.in_block(net)
        down_nets = []
        for down_block in self.down_blocks:
            net = down_block(net)
            down_nets.append(net)
        inner = self.last_down(net)
        up_nets = [inner]
        for down_net, up_block in zip(down_nets[::-1], self.up_blocks):
            if inner.shape != down_net.shape:
                inner = interpolate(
                    inner,
                    anchor=down_net,
                    mode="bilinear",
                    determinate=determinate,
                )
            inner = torch.cat([inner, down_net], dim=1)
            if not isinstance(up_block, nn.ModuleList):
                inner = up_block(inner)
            else:
                for block in up_block:
                    kw = {}
                    if isinstance(block, Interpolate):
                        kw["determinate"] = determinate
                    inner = block(inner, **kw)
            up_nets.append(inner)
        if self.return_up_nets:
            return up_nets
        return inner + in_net


class UNetRS(UNetBase):
    def __init__(
        self,
        in_channels: int = 3,
        mid_channels: int = 12,
        out_channels: int = 3,
        *,
        num_layers: int,
        inner_upsample_mode: str = "bilinear",
        inner_upsample_factor: Optional[int] = None,
    ):
        super().__init__()
        self.msg = f"{in_channels}, {mid_channels}, {out_channels}, {num_layers}"
        self.in_block = ConvSeq(in_channels, out_channels, dilation=1)
        blocks = [ConvSeq(out_channels, mid_channels, dilation=1)]
        for _ in range(num_layers - 2):
            blocks.append(
                nn.Sequential(
                    nn.MaxPool2d(2, stride=2),
                    ConvSeq(mid_channels, mid_channels, dilation=1),
                )
            )
        self.down_blocks = nn.ModuleList(blocks)
        self.last_down = ConvSeq(mid_channels, mid_channels, dilation=2)
        basic_block = ConvSeq(mid_channels * 2, mid_channels, dilation=1)
        if inner_upsample_factor is not None:
            basic_block = nn.ModuleList(
                [
                    basic_block,
                    Interpolate(inner_upsample_factor, inner_upsample_mode),
                ]
            )
        blocks = get_clones(basic_block, num_layers - 2, return_list=True)
        blocks.append(ConvSeq(mid_channels * 2, out_channels, dilation=1))
        self.up_blocks = nn.ModuleList(blocks)

    def extra_repr(self) -> str:
        return self.msg


class UNetFRS(UNetBase):
    def __init__(
        self,
        in_channels: int = 3,
        mid_channels: int = 12,
        out_channels: int = 3,
        *,
        num_layers: int,
    ):
        super().__init__()
        self.msg = f"{in_channels}, {mid_channels}, {out_channels}, {num_layers}"
        self.in_block = ConvSeq(in_channels, out_channels, dilation=1)
        blocks = [ConvSeq(out_channels, mid_channels, dilation=1)]
        for i in range(num_layers - 2):
            blocks.append(ConvSeq(mid_channels, mid_channels, dilation=2 ** (i + 1)))
        self.down_blocks = nn.ModuleList(blocks)
        self.last_down = ConvSeq(mid_channels, mid_channels, dilation=8)
        blocks = []
        for i in range(num_layers - 2):
            dilation = 2 ** (num_layers - i - 2)
            blocks.append(ConvSeq(mid_channels * 2, mid_channels, dilation=dilation))
        blocks.append(ConvSeq(mid_channels * 2, out_channels, dilation=1))
        self.up_blocks = nn.ModuleList(blocks)

    def extra_repr(self) -> str:
        return self.msg


@register_module("u2net")
class U2Net(UNetBase, ImageTranslatorMixin):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        upsample_mode: str = "bilinear",
        latent_channels: int = 32,
        num_layers: int = 5,
        num_inner_layers: int = 7,
        lite: bool = False,
    ):
        super().__init__()
        self.return_up_nets = True

        mid_nc = latent_channels // 2 if lite else latent_channels
        out_nc = latent_channels * 2
        in_nc = out_nc if lite else latent_channels
        current_layers = num_inner_layers

        self.in_block = nn.Identity()
        blocks = [
            UNetRS(
                in_channels,
                mid_nc,
                out_nc,
                num_layers=current_layers,
                inner_upsample_mode=upsample_mode,
                inner_upsample_factor=2,
            ),
        ]
        for i in range(num_layers - 2):
            if lite:
                in_nc = out_nc
            else:
                in_nc = out_nc
                out_nc *= 2
            current_layers = num_inner_layers - i - 1
            blocks.append(
                nn.Sequential(
                    nn.MaxPool2d(2, stride=2),
                    UNetRS(
                        in_nc,
                        mid_nc,
                        out_nc,
                        num_layers=current_layers,
                        inner_upsample_mode=upsample_mode,
                        inner_upsample_factor=2,
                    ),
                )
            )
            if not lite:
                mid_nc *= 2
        if not lite:
            in_nc *= 2
        blocks.append(
            nn.Sequential(
                nn.MaxPool2d(2, stride=2),
                UNetFRS(in_nc, mid_nc, out_nc, num_layers=current_layers),
            )
        )
        self.down_blocks = nn.ModuleList(blocks)
        self.last_down = nn.Sequential(
            nn.MaxPool2d(2, stride=2),
            UNetFRS(in_nc, mid_nc, out_nc, num_layers=current_layers),
        )
        in_nc *= 2
        ncs = [out_nc] * 2
        blocks = [UNetFRS(in_nc, mid_nc, out_nc, num_layers=current_layers)]
        for i in range(num_layers - 2):
            if not lite:
                mid_nc //= 2
                out_nc //= 2
            blocks.append(
                UNetRS(
                    in_nc,
                    mid_nc,
                    out_nc,
                    num_layers=current_layers,
                    inner_upsample_mode=upsample_mode,
                    inner_upsample_factor=2,
                )
            )
            current_layers += 1
            if not lite:
                in_nc //= 2
            ncs.append(out_nc)
        if not lite:
            mid_nc //= 2
        blocks.append(
            UNetRS(
                in_nc,
                mid_nc,
                out_nc,
                num_layers=current_layers,
                inner_upsample_mode=upsample_mode,
                inner_upsample_factor=2,
            )
        )
        ncs.append(out_nc)
        self.up_blocks = nn.ModuleList(blocks)

        blocks = []
        for i, nc in enumerate(ncs[::-1]):
            blocks.append(Conv2d(nc, out_channels, kernel_size=3, padding=1))
        self.side_blocks = nn.ModuleList(blocks)
        self.out = Conv2d((len(ncs)) * out_channels, out_channels, kernel_size=1)

    def forward(self, net: Tensor, *, determinate: bool = False) -> List[Tensor]:
        up_nets = super().forward(net)
        side_nets: List[Tensor] = []
        for up_net, side_block in zip(up_nets[::-1], self.side_blocks):
            side_net = side_block(up_net)
            if side_nets and side_net.shape != side_nets[0].shape:
                side_net = interpolate(
                    side_net,
                    anchor=side_nets[0],
                    mode="bilinear",
                    determinate=determinate,
                )
            side_nets.append(side_net)
        side_nets.insert(0, self.out(torch.cat(side_nets, dim=1)))
        return side_nets


__all__ = ["U2Net"]



import os

import numpy as np

from typing import List
from typing import Tuple
from typing import Optional
from dataclasses import dataclass
from cftool.misc import print_warning
from cftool.array import to_numpy
from cftool.array import to_torch

from ..pipeline import MLPipeline
from ....misc.toolkit import eval_context
from ....models.ml.nbm import NBM
from ....models.protocols.ml import IEncoder
from ....models.protocols.ml import Dimensions

try:
    import matplotlib.pyplot as plt
except:
    plt = None


@dataclass
class NBMInspector:
    m: NBM
    encoder: Optional[IEncoder] = None
    dimensions: Optional[Dimensions] = None
    dpi: int = 200
    figsize: Tuple[int, int] = (8, 6)
    boundings: Optional[List[Tuple[float, float]]] = None
    column_names: Optional[List[str]] = None

    def fit(
        self,
        x: np.ndarray,
        column_names: Optional[List[str]] = None,
    ) -> "NBMInspector":
        self.boundings = []
        for i in range(x.shape[-1]):
            column = x[..., i]
            self.boundings.append((column.min().item(), column.max().item()))
        self.column_names = column_names
        return self

    def visualize(
        self,
        target_x_dims: List[Tuple[int, ...]],
        y_dim: int,
        export_folder: str,
        *,
        dense: int = 400,
        device: str = "cpu",
    ) -> None:
        if plt is None:
            raise ValueError("`matplotlib` is needed for `visualize`")
        if self.boundings is None:
            msg = "`boundings` is not defined, did you forget to call `fit`?"
            raise ValueError(msg)
        os.makedirs(export_folder, exist_ok=True)
        self.m.to(device)
        existing_titles = set()
        for x_dims in target_x_dims:
            if len(x_dims) != 1:
                print_warning("dim > 1 is not supported yet")
                continue

            if self.column_names is None:
                title = "-".join(str(i) for i in sorted(x_dims))
            else:
                title = "-".join(self.column_names[i] for i in sorted(x_dims))
            title = f"{title} ({y_dim})"
            if title in existing_titles:
                continue
            existing_titles.add(title)

            # currently only support one dim
            dim = x_dims[0]
            if self.dimensions is None:
                indices = x_dims
                is_categorical = False
            else:
                res = self.dimensions.get_indices_in_merged(dim)
                if res is None:
                    print_warning(f"column {dim} is redundant, skipped")
                    continue
                indices, is_categorical = res
            i_min, i_max = self.boundings[dim]
            if not is_categorical:
                x = np.random.random([dense, 1]) * (i_max - i_min) + i_min
                x = x[np.argsort(x.ravel())]
                net = to_torch(x).to(device)
                net = self.m.inspect(net, indices, y_dim, already_extracted=True)
                response = to_numpy(net)
                self._export_numerical(
                    x,
                    response,
                    title,
                    os.path.join(export_folder, f"{title}.png"),
                )
            else:
                if self.encoder is None:
                    print(
                        f"column {dim} is categorical "
                        "but `encoder` is not provided, skipped"
                    )
                    continue
                x = np.arange(i_min, i_max + 1)[..., None]
                with eval_context(self.encoder):
                    encoded = self.encoder.encode(x, dim)
                net = encoded.merged
                response = np.zeros([net.shape[0], 1], np.float32)
                for i, idx in enumerate(indices):
                    i_net = net[..., [i]]
                    i_pack = (idx,)
                    i_net = self.m.inspect(i_net, i_pack, y_dim, already_extracted=True)
                    response += to_numpy(i_net)
                self._export_categorical(
                    len(x),
                    response,
                    title,
                    os.path.join(export_folder, f"{title}.png"),
                )

    def _export_numerical(
        self,
        x: np.ndarray,
        response: np.ndarray,
        title: str,
        export_path: str,
    ) -> None:
        fig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        plt.title(title)
        plt.plot(x.ravel(), response.ravel(), label="response")
        plt.legend()
        fig.savefig(export_path)
        plt.close()

    def _export_categorical(
        self,
        n: int,
        response: np.ndarray,
        title: str,
        export_path: str,
    ) -> None:
        x_base = np.arange(1, n + 1)
        response = response.ravel()
        y_min, y_max = response.min(), response.max()
        padding = 0.1 * (y_max - y_min)
        fig = plt.figure(figsize=self.figsize, dpi=self.dpi)
        plt.title(title)
        plt.bar(
            x_base,
            response,
            width=0.35,
            edgecolor="white",
        )
        plt.xticks(
            [i for i in range(n + 2)],
            [""] + [str(i) for i in range(n)] + [""],
        )
        plt.setp(plt.xticks()[1], rotation=30, horizontalalignment="right")
        plt.ylim(y_min - padding, y_max + padding)
        fig.tight_layout()
        fig.savefig(export_path)
        plt.close()

    @classmethod
    def from_pipeline(
        cls,
        m: MLPipeline,
        dpi: int = 200,
        figsize: Tuple[int, int] = (8, 6),
    ) -> "NBMInspector":
        model = m.model
        if isinstance(model.dimensions, list):
            raise ValueError(
                "`MLPipeline` with `_num_repeat` defined is detected, "
                "which is not compatible with `NBM.from_pipeline`"
            )
        return cls(
            model.core.core,
            model.encoder,
            model.dimensions,
            dpi,
            figsize,
        )


__all__ = [
    "NBMInspector",
]



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional
from cftool.array import squeeze

from .protocol import EncoderMixin
from .protocol import Encoder1DMixin
from ....modules.blocks import get_conv_blocks
from ....modules.blocks import Linear
from ....modules.blocks import ResidualBlock


# TODO : Try ResidualBlockV2


@EncoderMixin.register("vanilla")
class VanillaEncoder(nn.Module, EncoderMixin):
    def __init__(
        self,
        in_channels: int,
        num_downsample: int,
        latent_channels: int = 256,
        *,
        kernel_size: int = 3,
        first_kernel_size: int = 7,
        start_channels: Optional[int] = None,
        num_residual_blocks: int = 0,
        residual_dropout: float = 0.0,
        residual_kwargs: Optional[Dict[str, Any]] = None,
        norm_type: Optional[str] = "batch",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        activation: str = "leaky_relu_0.2",
        padding: str = "same",
    ):
        super().__init__()
        self.in_channels = in_channels
        self.num_downsample = num_downsample
        self.latent_channels = latent_channels
        self.first_kernel_size = first_kernel_size
        if start_channels is None:
            start_channels = int(round(latent_channels / (2**num_downsample)))
        if start_channels <= 0:
            raise ValueError(
                f"latent_channels ({latent_channels}) is too small "
                f"for num_downsample ({num_downsample})"
            )
        # in conv
        blocks = get_conv_blocks(
            in_channels,
            start_channels,
            first_kernel_size,
            1,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
            activation=activation,
            padding=padding,
        )
        # downsample
        downsample_padding = padding
        if padding == "reflection":
            downsample_padding = f"reflection{(kernel_size - 1) // 2}"
        in_nc = start_channels
        for i in range(self.num_downsample):
            is_last = i == self.num_downsample - 1
            i_norm_type = norm_type
            i_activation = activation
            if not is_last:
                out_nc = min(in_nc * 2, latent_channels)
            else:
                out_nc = latent_channels
                if num_residual_blocks == 0:
                    i_norm_type = None
                    i_activation = None  # type: ignore
            blocks.extend(
                get_conv_blocks(
                    in_nc,
                    out_nc,
                    kernel_size,
                    2,
                    norm_type=i_norm_type,
                    activation=i_activation,
                    padding=downsample_padding,
                )
            )
            in_nc = out_nc
        # residual
        for _ in range(num_residual_blocks):
            blocks.append(
                ResidualBlock(
                    latent_channels,
                    residual_dropout,
                    norm_type=norm_type,
                    activation=activation,
                    padding=padding,
                    **(residual_kwargs or {}),
                )
            )
        # construct
        self.encoder = nn.Sequential(*blocks)

    def forward(self, net: Tensor) -> Tensor:
        return self.encoder(net)


@Encoder1DMixin.register("vanilla")
class VanillaEncoder1D(nn.Module, Encoder1DMixin):
    def __init__(
        self,
        in_channels: int,
        num_downsample: int,
        latent_dim: int = 128,
        *,
        img_size: Optional[int] = None,
        kernel_size: int = 3,
        first_kernel_size: int = 7,
        start_channels: Optional[int] = None,
        num_residual_blocks: int = 0,
        residual_dropout: float = 0.0,
        residual_kwargs: Optional[Dict[str, Any]] = None,
        norm_type: Optional[str] = "batch",
        activation: str = "leaky_relu_0.2",
        padding: str = "same",
        pool: str = "average",
    ):
        super().__init__()
        self.in_channels = in_channels
        self.latent_dim = latent_dim
        self.encoder = VanillaEncoder(
            in_channels,
            num_downsample,
            latent_dim,
            kernel_size=kernel_size,
            first_kernel_size=first_kernel_size,
            start_channels=start_channels,
            num_residual_blocks=num_residual_blocks,
            residual_dropout=residual_dropout,
            residual_kwargs=residual_kwargs,
            norm_type=norm_type,
            activation=activation,
            padding=padding,
        )
        if pool == "average":
            self.pool = nn.AdaptiveAvgPool2d((1, 1))
        elif pool == "fc":
            if img_size is None:
                raise ValueError("`img_size` should be provided if `pool`=fc")
            res = self.encoder.latent_resolution(img_size)
            flattened_dim = latent_dim * res**2
            self.pool = nn.Sequential(nn.Flatten(1), Linear(flattened_dim, latent_dim))
        else:
            raise ValueError(f"unrecognized `pool` value : '{pool}'")

    def forward(self, net: Tensor) -> Tensor:
        net = self.encoder(net)
        net = squeeze(self.pool(net))
        return net


__all__ = [
    "VanillaEncoder",
    "VanillaEncoder1D",
]



from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional

from .monitors import BasicMonitor
from ...types import configs_type
from ...trainer import Trainer
from ...protocol import TqdmSettings
from ...protocol import OptimizerPack
from ...protocol import TrainerMonitor
from ...protocol import _IMetric
from ...protocol import _MultipleMetrics
from ...protocol import TrainerCallback
from .callbacks.general import _LogMetricsMsgCallback


def make_trainer(
    state_config: Optional[Dict[str, Any]] = None,
    *,
    workplace: str,
    num_epoch: int = 40,
    max_epoch: int = 1000,
    fixed_epoch: Optional[int] = None,
    fixed_steps: Optional[int] = None,
    log_steps: Optional[int] = None,
    valid_portion: float = 1.0,
    amp: bool = False,
    clip_norm: float = 0.0,
    metric_names: Optional[Union[str, List[str]]] = None,
    metric_configs: configs_type = None,
    metric_weights: Optional[Dict[str, float]] = None,
    use_losses_as_metrics: Optional[bool] = None,
    loss_metrics_weights: Optional[Dict[str, float]] = None,
    recompute_train_losses_in_eval: bool = True,
    monitor_names: Optional[Union[str, List[str]]] = None,
    monitor_configs: configs_type = None,
    callback_names: Optional[Union[str, List[str]]] = None,
    callback_configs: configs_type = None,
    lr: Optional[float] = None,
    optimizer_name: Optional[str] = None,
    scheduler_name: Optional[str] = None,
    optimizer_config: Optional[Dict[str, Any]] = None,
    scheduler_config: Optional[Dict[str, Any]] = None,
    update_scheduler_per_epoch: bool = False,
    optimizer_settings: Optional[Dict[str, Dict[str, Any]]] = None,
    use_zero: bool = False,
    data_info_name: str = "data_info",
    metrics_log_file: str = "metrics.txt",
    finetune_config: Optional[Dict[str, Any]] = None,
    tqdm_settings: Optional[Dict[str, Any]] = None,
) -> Trainer:
    if state_config is None:
        state_config = {}
    if log_steps is not None:
        state_config.setdefault("num_step_per_log", log_steps)
        state_config.setdefault("snapshot_start_step", log_steps)
        state_config.setdefault("num_step_per_snapshot", log_steps)
    # metrics
    metrics: Optional[Union[_IMetric, _MultipleMetrics]]
    if metric_names is None:
        metrics = None
    else:
        metrics = _IMetric.fuse(
            metric_names,
            metric_configs,
            metric_weights=metric_weights,
        )
    # monitor
    monitors: Optional[List[TrainerMonitor]]
    if monitor_names is None:
        monitors = None
    else:
        monitors = TrainerMonitor.make_multiple(monitor_names, monitor_configs)
    # callback
    callbacks: Optional[List[TrainerCallback]]
    if callback_names is None:
        callbacks = None
    else:
        callbacks = TrainerCallback.make_multiple(callback_names, callback_configs)
    # optimizer
    if optimizer_settings is None:
        optimizer_packs = None
    else:
        optimizer_packs = []
        for key, settings in optimizer_settings.items():
            optimizer = settings.get("optimizer")
            if optimizer is None:
                raise ValueError(f"optimizer must be provided (key={key})")
            optimizer_packs.append(
                OptimizerPack(
                    key,
                    optimizer,
                    settings.get("scheduler"),
                    settings.get("optimizer_config"),
                    settings.get("scheduler_config"),
                )
            )
    # tqdm
    if tqdm_settings is None:
        tqdm_settings = {}
    use_tqdm = tqdm_settings.setdefault("use_tqdm", False)
    # epoch
    if fixed_epoch is not None:
        num_epoch = max_epoch = fixed_epoch
    if max_epoch < num_epoch:
        raise ValueError("`max_epoch` should not be smaller than `num_epoch`")
    # default behaviors
    if monitors is None:
        monitors = [BasicMonitor()]
    if callbacks is None:
        callbacks = [_LogMetricsMsgCallback(not use_tqdm)]
    return Trainer(
        state_config,
        num_epoch=num_epoch,
        max_epoch=max_epoch,
        fixed_steps=fixed_steps,
        valid_portion=valid_portion,
        amp=amp,
        clip_norm=clip_norm,
        metrics=metrics,
        use_losses_as_metrics=use_losses_as_metrics,
        loss_metrics_weights=loss_metrics_weights,
        recompute_train_losses_in_eval=recompute_train_losses_in_eval,
        monitors=monitors,
        callbacks=callbacks,
        lr=lr,
        optimizer_name=optimizer_name,
        scheduler_name=scheduler_name,
        optimizer_config=optimizer_config,
        scheduler_config=scheduler_config,
        update_scheduler_per_epoch=update_scheduler_per_epoch,
        optimizer_packs=optimizer_packs,
        use_zero=use_zero,
        workplace=workplace,
        data_info_name=data_info_name,
        metrics_log_file=metrics_log_file,
        finetune_config=finetune_config,
        tqdm_settings=TqdmSettings(
            use_tqdm,
            tqdm_settings.setdefault("use_step_tqdm", use_tqdm),
            tqdm_settings.setdefault("use_tqdm_in_validation", False),
            tqdm_settings.setdefault("in_distributed", False),
            tqdm_settings.setdefault("tqdm_position", 0),
            tqdm_settings.setdefault("tqdm_desc", "epoch"),
        ),
    )


__all__ = [
    "make_trainer",
]



import torch

from torch import nn
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional
from cftool.types import tensor_dict_type

from ....protocol import TrainerState
from ....protocol import WithDeviceMixin
from ....constants import PREDICTIONS_KEY
from ..generator.vector_quantized import VQGenerator
from ....misc.toolkit import auto_num_layers
from ....modules.blocks import ApplyTanhMixin
from ....misc.internal_.register import register_module


@register_module("vq_vae")
class VQVAE(nn.Module, WithDeviceMixin, ApplyTanhMixin):
    def __init__(
        self,
        img_size: int,
        num_code: int,
        in_channels: int = 3,
        out_channels: Optional[int] = None,
        num_downsample: Optional[int] = None,
        min_size: int = 8,
        target_downsample: int = None,
        *,
        encoder: str = "vanilla",
        decoder: str = "vanilla",
        code_dimension: int = 256,
        latent_channels: int = 256,
        encoder_config: Optional[Dict[str, Any]] = None,
        decoder_config: Optional[Dict[str, Any]] = None,
        latent_padding_channels: Optional[int] = None,
        num_classes: Optional[int] = None,
        apply_tanh: bool = True,
    ):
        super().__init__()
        self.img_size = img_size
        self.num_classes = num_classes
        self.latent_channels = latent_channels
        self.apply_tanh = apply_tanh
        if num_downsample is None:
            args = img_size, min_size, target_downsample
            num_downsample = auto_num_layers(*args, use_stride=encoder == "vanilla")
        # encoder
        if encoder_config is None:
            encoder_config = {}
        encoder_config["num_downsample"] = num_downsample
        # decoder
        if decoder_config is None:
            decoder_config = {}
        decoder_config["num_upsample"] = num_downsample
        # vq generator
        self.generator = VQGenerator(
            img_size,
            num_code,
            in_channels,
            out_channels,
            encoder=encoder,
            decoder=decoder,
            code_dimension=code_dimension,
            latent_channels=latent_channels,
            encoder_config=encoder_config,
            decoder_config=decoder_config,
            latent_padding_channels=latent_padding_channels,
            num_classes=num_classes,
        )
        self.latent_resolution = self.generator.latent_resolution

    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        apply_tanh: Optional[bool] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        results = self.generator(batch_idx, batch, state, return_z_q_g=True, **kwargs)
        net = results[PREDICTIONS_KEY]
        results[PREDICTIONS_KEY] = self.postprocess(net, apply_tanh)
        return results

    def decode(
        self,
        z_q: Tensor,
        *,
        labels: Optional[Tensor] = None,
        apply_tanh: Optional[bool] = None,
        resize: bool = True,
    ) -> Tensor:
        net = self.generator.decode(z_q, labels=labels, resize=resize)
        net = self.postprocess(net, apply_tanh)
        return net

    def get_code_indices(self, net: Tensor, **kwargs: Any) -> Tensor:
        return self.generator.get_code_indices(net, **kwargs)

    def get_code(self, code_indices: Tensor) -> Tensor:
        return self.generator.get_code(code_indices)

    def reconstruct_from(
        self,
        code_indices: Tensor,
        *,
        class_idx: Optional[int] = None,
        labels: Optional[Tensor] = None,
        use_one_hot: bool = False,
        apply_tanh: Optional[bool] = None,
        **kwargs: Any,
    ) -> Tensor:
        net = self.generator.reconstruct_from(
            code_indices,
            class_idx=class_idx,
            labels=labels,
            use_one_hot=use_one_hot,
            **kwargs,
        )
        net = self.postprocess(net, apply_tanh)
        return net

    def sample_codebook(
        self,
        *,
        code_indices: Optional[Tensor] = None,
        num_samples: Optional[int] = None,
        class_idx: Optional[int] = None,
        **kwargs: Any,
    ) -> Tuple[Tensor, Tensor]:
        return self.generator.sample_codebook(
            code_indices=code_indices,
            num_samples=num_samples,
            class_idx=class_idx,
            **kwargs,
        )


__all__ = ["VQVAE"]



import numpy as np

from PIL import Image
from typing import Any
from typing import Dict
from cftool.array import arr_type
from cftool.array import to_torch
from torchvision.transforms import transforms

from ..basic import Transforms


class NoBatchTransforms(Transforms):
    @property
    def need_batch_process(self) -> bool:
        return False


class BatchWrapper(Transforms):
    def __init__(self, transform: Transforms, input_alias: str):
        super().__init__()
        self.transform = transform
        self.input_alias = input_alias

    def fn(self, inp: Dict[str, Any]) -> Dict[str, Any]:
        return {
            key: value if key != self.input_alias else self.transform(value)
            for key, value in inp.items()
        }

    @property
    def need_numpy(self) -> bool:
        return self.transform.need_numpy

    @property
    def need_batch_process(self) -> bool:
        return True


@Transforms.register("to_array")
class ToArray(Transforms):
    def __init__(self, need_batch_process: bool = False):
        super().__init__()
        self._batch = need_batch_process

    @staticmethod
    def to_array(v: Any) -> np.ndarray:
        return np.asarray(v).astype(np.float32) / 255.0

    def fn(self, inp: Any) -> Any:
        if not self._batch:
            return self.to_array(inp)
        return {k: self.to_array(v) for k, v in inp.items()}

    @property
    def need_numpy(self) -> bool:
        return False

    @property
    def need_batch_process(self) -> bool:
        return self._batch


@Transforms.register("to_rgb")
class ToRGB(NoBatchTransforms):
    @staticmethod
    def arr_fn(img: arr_type) -> arr_type:
        is_numpy = isinstance(img, np.ndarray)
        if len(img.shape) == 2:
            img = img[..., None] if is_numpy else img[None, ...]
        c = img.shape[2 if is_numpy else 0]
        if c == 3:
            return img
        if c == 4:
            if is_numpy:
                return img[..., :3] * img[..., 3:]
            return img[:3] * img[3:]
        if c == 1:
            if is_numpy:
                return img.repeat(3, axis=2)
            return img.repeat([3, 1, 1])
        raise ValueError(f"invalid shape occurred ({img.shape})")

    def fn(self, inp: Any) -> Any:
        if isinstance(inp, Image.Image):
            inp = np.array(inp)
        return self.arr_fn(inp)

    @property
    def need_numpy(self) -> bool:
        return False


@Transforms.register("to_gray")
class ToGray(NoBatchTransforms):
    def __init__(self):  # type: ignore
        super().__init__()
        self.pt_fn = transforms.Grayscale()

    def np_fn(self, img: np.ndarray) -> np.ndarray:
        if len(img.shape) == 2:
            return img[..., None]
        if img.shape[2] == 1:
            return img
        tensor = to_torch(img.transpose([2, 0, 1]))
        return self.pt_fn(tensor).contiguous().numpy().transpose([1, 2, 0])

    def fn(self, inp: Any) -> Any:
        if isinstance(inp, np.ndarray):
            return self.np_fn(inp)
        return self.pt_fn(inp)

    @property
    def need_numpy(self) -> bool:
        return False


__all__ = [
    "ToRGB",
    "ToGray",
    "ToArray",
    "BatchWrapper",
    "NoBatchTransforms",
]



from torch import nn
from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional

from ..decoder.vanilla import VanillaDecoder
from ..encoder.vanilla import VanillaEncoder
from ...protocols.cv import ImageTranslatorMixin
from ....constants import INPUT_KEY
from ....misc.toolkit import interpolate
from ....misc.internal_.register import register_module


@register_module("cycle_gan_generator")
class CycleGANGenerator(nn.Module, ImageTranslatorMixin):
    def __init__(
        self,
        in_channels: int = 3,
        out_channels: Optional[int] = None,
        num_downsample: int = 2,
        *,
        kernel_size: int = 3,
        first_kernel_size: int = 7,
        start_channels: int = 64,
        num_residual_blocks: int = 9,
        residual_dropout: float = 0.0,
        norm_type: Optional[str] = "instance",
        norm_kwargs: Optional[Dict[str, Any]] = None,
        activation: str = "relu",
        padding: str = "reflection",
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels or in_channels
        self.num_downsample = num_downsample
        self.start_channels = start_channels
        self.latent_channels = start_channels * 2**num_downsample
        self.encoder = VanillaEncoder(
            in_channels,
            num_downsample,
            self.latent_channels,
            kernel_size=kernel_size,
            first_kernel_size=first_kernel_size,
            start_channels=self.start_channels,
            num_residual_blocks=num_residual_blocks,
            residual_dropout=residual_dropout,
            norm_type=norm_type,
            norm_kwargs=norm_kwargs,
            activation=activation,
            padding=padding,
        )
        self.decoder = VanillaDecoder(
            self.latent_channels,
            self.out_channels,
            norm_type=norm_type,
            activation=activation,
            padding=padding,
            kernel_size=kernel_size,
            last_kernel_size=first_kernel_size,
            num_repeats="default",
            reduce_channel_on_upsample=True,
            num_upsample=num_downsample,
        )

    def forward(self, inp: Tensor) -> Tensor:
        net = self.encoder(inp)
        net = self.decoder({INPUT_KEY: net})
        net = interpolate(net, anchor=inp)
        return net


__all__ = ["CycleGANGenerator"]



from torch import nn
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Callable
from typing import Optional
from cftool.array import squeeze
from cftool.types import tensor_dict_type

from .core import Backbone
from ..protocol import EncoderMixin
from ..protocol import Encoder1DMixin
from .....constants import LATENT_KEY
from .....modules.blocks import Conv2d


class Preset:
    remove_layers: Dict[str, List[str]] = {}
    target_layers: Dict[str, Dict[str, str]] = {}
    increment_configs: Dict[str, Dict[str, Any]] = {}

    @classmethod
    def register_settings(cls) -> Callable:
        def _register(settings: Preset) -> None:
            cls.remove_layers.update(settings.remove_layers)
            cls.target_layers.update(settings.target_layers)
            cls.increment_configs.update(settings.increment_configs)

        return _register


@EncoderMixin.register("backbone")
class BackboneEncoder(nn.Module, EncoderMixin):
    def __init__(
        self,
        name: str,
        in_channels: int,
        *,
        finetune: bool = True,
        pretrained: bool = False,
        use_to_rgb: bool = False,
        to_rgb_bias: bool = False,
        backbone_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.in_channels = in_channels
        if in_channels == 3 and not use_to_rgb:
            self.to_rgb = None
        else:
            self.to_rgb = Conv2d(in_channels, 3, kernel_size=1, bias=to_rgb_bias)
        self.net = Backbone(
            name,
            pretrained=pretrained,
            requires_grad=finetune,
            **(backbone_config or {}),
        )
        self.num_downsample = self.net.num_downsample
        self.latent_channels = self.net.latent_channels

    def forward(self, net: Tensor) -> tensor_dict_type:
        if self.to_rgb is not None:
            net = self.to_rgb(net)
        return self.net(net)


@Encoder1DMixin.register("backbone")
class BackboneEncoder1D(nn.Module, Encoder1DMixin):
    def __init__(
        self,
        name: str,
        in_channels: int,
        *,
        finetune: bool = True,
        pretrained: bool = False,
        backbone_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.in_channels = in_channels
        self.encoder = BackboneEncoder(
            name,
            in_channels,
            finetune=finetune,
            pretrained=pretrained,
            backbone_config=backbone_config,
        )
        self.latent_dim = self.encoder.latent_channels
        self.pool = nn.AdaptiveAvgPool2d((1, 1))

    def forward(self, net: Tensor) -> tensor_dict_type:
        outputs = self.encoder(net)
        latent = outputs[LATENT_KEY]
        if latent.shape[-2] != 1 or latent.shape[-1] != 1:
            latent = self.pool(latent)
        outputs[LATENT_KEY] = squeeze(latent)
        return outputs


__all__ = [
    "BackboneEncoder",
    "BackboneEncoder1D",
]
